@@ -5,12 +5,14 @@
 // License           : BSD License
 // ----------------------------------------------------------------------------
 // Creation date     : 12-Dec-2014
-// Last modification : 22-Dec-2014
+// Last modification : 30-Dec-2014
 // ----------------------------------------------------------------------------
 
 #include "parsing.h"
 #include "oo.h"
 
+#define MATCH_STATS(m) ParsingStats_registerMatch(context->stats, this, m)
+
 // SEE: https://en.wikipedia.org/wiki/C_data_types
 // SEE: http://stackoverflow.com/questions/18329532/pcre-is-not-matching-utf8-characters
 // HASH: search.h, hcreate, hsearch, etc
@@ -110,10 +112,11 @@ bool Iterator_moveTo ( Iterator* this, size_t offset ) {
 	return this->move(this, offset - this->offset );
 }
 
-
 void Iterator_free( Iterator* this ) {
-	// TODO: Take care of input
+	// FIXME: Should close input file
+	__DEALLOC(this->buffer);
 	__DEALLOC(this);
+
 }
 
 // ----------------------------------------------------------------------------
@@ -281,9 +284,22 @@ Grammar* Grammar_new(void) {
 	this->skip       = NULL;
 	this->axiomCount = 0;
 	this->skipCount  = 0;
+	this->elements   = NULL;
+	this->isVerbose  = FALSE;
 	return this;
 }
 
+void Grammar_free(Grammar* this) {
+	// FIXME: Implement me. We should get a list of all the parsing elements
+	// and free them.
+	/*
+	Element*[] symbols = Grammar_listSymbols();
+	for (int i = 0; i < (this->axiomCount + this->skipCount); i++ ) {
+	}
+	*/
+	__DEALLOC(this);
+}
+
 // ----------------------------------------------------------------------------
 //
 // MATCH
@@ -327,13 +343,13 @@ bool Match_isSuccess(Match* this) {
 	return (this != NULL && this != FAILURE && this->status == STATUS_MATCHED);
 }
 
-int Match__walk(Match* this, WalkingCallback callback, int step ){
-	step = callback(this, step);
+int Match__walk(Match* this, WalkingCallback callback, int step, void* context ){
+	step = callback(this, step, context);
 	if (this->child != NULL && step >= 0) {
-		step = Match__walk(this->child, callback, step + 1);
+		step = Match__walk(this->child, callback, step + 1, context);
 	}
 	if (this->next != NULL && step >= 0) {
-		step = Match__walk(this->next, callback, step + 1);
+		step = Match__walk(this->next, callback, step + 1, context);
 	}
 	return step;
 }
@@ -418,12 +434,12 @@ ParsingElement* ParsingElement_name( ParsingElement* this, const char* name ) {
 	return this;
 }
 
-int ParsingElement__walk( ParsingElement* this, WalkingCallback callback, int step ) {
+int ParsingElement__walk( ParsingElement* this, WalkingCallback callback, int step, void* context ) {
 	int i = step;
-	step  = callback((Element*)this, step);
+	step  = callback((Element*)this, step, context);
 	Reference* child = this->children;
 	while ( child != NULL && step >= 0) {
-		int j = Reference__walk(child, callback, ++i);
+		int j = Reference__walk(child, callback, ++i, context);
 		if (j > 0) { step = i = j; }
 		child = child->next;
 	}
@@ -436,20 +452,20 @@ int ParsingElement__walk( ParsingElement* this, WalkingCallback callback, int st
 //
 // ----------------------------------------------------------------------------
 
-int Element_walk( Element* this, WalkingCallback callback ) {
-	return Element__walk(this, callback, 0);
+int Element_walk( Element* this, WalkingCallback callback, void* context ) {
+	return Element__walk(this, callback, 0, context);
 }
 
-int Element__walk( Element* this, WalkingCallback callback, int step ) {
+int Element__walk( Element* this, WalkingCallback callback, int step, void* context ) {
 	assert (callback != NULL);
 	if (this!=NULL) {
 		if (Reference_Is(this)) {
-			step = Reference__walk((Reference*)this, callback, step);
+			step = Reference__walk((Reference*)this, callback, step, context);
 		} else if (ParsingElement_Is(this)) {
-			step = ParsingElement__walk((ParsingElement*)this, callback, step);
+			step = ParsingElement__walk((ParsingElement*)this, callback, step, context);
 		} else {
 			// If it is neither a reference or parsing element, it is a Match
-			step = Match__walk((Match*)this, callback, step);
+			step = Match__walk((Match*)this, callback, step, context);
 		}
 	}
 	return step;
@@ -503,10 +519,10 @@ Reference* Reference_name(Reference* this, const char* name) {
 	return this;
 }
 
-int Reference__walk( Reference* this, WalkingCallback callback, int step ) {
-	step = callback((Element*)this, step);
+int Reference__walk( Reference* this, WalkingCallback callback, int step, void* context ) {
+	step = callback((Element*)this, step, context);
 	if (step >= 0) {
-		step = ParsingElement__walk(this->element, callback, step + 1);
+		step = ParsingElement__walk(this->element, callback, step + 1, context);
 	}
 	return step;
 }
@@ -560,7 +576,7 @@ Match* Reference_recognize(Reference* this, ParsingContext* context) {
 		default:
 			// Unsuported cardinality
 			ERROR("Unsupported cardinality %c", this->cardinality);
-			return FAILURE;
+			return MATCH_STATS(FAILURE);
 	}
 	if (Match_isSuccess(result)) {
 		int    length   = context->iterator->offset - offset;
@@ -568,9 +584,9 @@ Match* Reference_recognize(Reference* this, ParsingContext* context) {
 		assert(result->element != NULL);
 		m->child        = result;
 		m->offset       = offset;
-		return m;
+		return MATCH_STATS(m);
 	} else {
-		return FAILURE;
+		return MATCH_STATS(FAILURE);
 	}
 }
 
@@ -602,11 +618,11 @@ Match* Word_recognize(ParsingElement* this, ParsingContext* context) {
 		// NOTE: You can see here that the word actually consumes input
 		// and moves the iterator.
 		context->iterator->move(context->iterator, config->length);
-		DEBUG("[✓] %s:%s matched at %zd", this->name, ((WordConfig*)this->config)->word, context->iterator->offset);
-		return Match_Success(config->length, this, context);
+		LOG_IF(context->grammar->isVerbose, "[✓] %s:%s matched at %zd", this->name, ((WordConfig*)this->config)->word, context->iterator->offset);
+		return MATCH_STATS(Match_Success(config->length, this, context));
 	} else {
-		DEBUG("    %s:%s failed at %zd", this->name, ((WordConfig*)this->config)->word, context->iterator->offset);
-		return FAILURE;
+		LOG_IF(context->grammar->isVerbose, "    %s:%s failed at %zd", this->name, ((WordConfig*)this->config)->word, context->iterator->offset);
+		return MATCH_STATS(FAILURE);
 	}
 }
 
@@ -622,9 +638,10 @@ ParsingElement* Token_new(const char* expr) {
 	this->type           = TYPE_TOKEN;
 	this->recognize      = Token_recognize;
 	this->freeMatch      = TokenMatch_free;
+	config->expr   = expr;
+#ifdef WITH_PCRE
 	const char* pcre_error;
 	int         pcre_error_offset = -1;
-	config->expr   = expr;
 	config->regexp = pcre_compile(expr, PCRE_UTF8, &pcre_error, &pcre_error_offset, NULL);
 	if (pcre_error != NULL) {
 		ERROR("Token: cannot compile regular expression `%s` at %d: %s", expr, pcre_error_offset, pcre_error);
@@ -639,6 +656,7 @@ ParsingElement* Token_new(const char* expr) {
 		__DEALLOC(this);
 		return NULL;
 	}
+#endif
 	this->config = config;
 	return this;
 }
@@ -647,8 +665,10 @@ void Token_free(ParsingElement* this) {
 	TokenConfig* config = (TokenConfig*)this->config;
 	if (config != NULL) {
 		// FIXME: Not sure how to free a regexp
+#ifdef WITH_PCRE
 		if (config->regexp != NULL) {}
 		if (config->extra  != NULL) {pcre_free_study(config->extra);}
+#endif
 		__DEALLOC(config);
 	}
 	__DEALLOC(this);
@@ -658,6 +678,8 @@ Match* Token_recognize(ParsingElement* this, ParsingContext* context) {
 	assert(this->config);
 	if(this->config == NULL) {return FAILURE;}
 	TokenConfig* config     = (TokenConfig*)this->config;
+	Match* result           = NULL;
+#ifdef WITH_PCRE
 	// NOTE: This has to be a multiple of 3, according to `man pcre_exec`
 	int vector_length = 30;
 	int vector[vector_length];
@@ -671,7 +693,6 @@ Match* Token_recognize(ParsingElement* this, ParsingContext* context) {
 		PCRE_ANCHORED,                     // OPTIONS -- we do not skip position
 		vector,                            // Vector of matching offsets
 		vector_length);                    // Number of elements in the vector
-	Match* result = NULL;
 	if (r <= 0) {
 		// DEBUG("Token: %s FAILED on %s", config->expr, context->iterator->buffer);
 		switch(r) {
@@ -683,7 +704,7 @@ Match* Token_recognize(ParsingElement* this, ParsingContext* context) {
 			case PCRE_ERROR_NOMEMORY     : ERROR("Token:%s Ran out of memory", config->expr);                       break;
 			default                      : ERROR("Token:%s Unknown error", config->expr);                           break;
 		};
-		DEBUG("    %s:%s failed at %zd", this->name, config->expr, context->iterator->offset);
+		LOG_IF(context->grammar->isVerbose, "    %s:%s failed at %zd", this->name, config->expr, context->iterator->offset);
 	} else {
 		if(r == 0) {
 			ERROR("Token: %s many substrings matched\n", config->expr);
@@ -696,7 +717,7 @@ Match* Token_recognize(ParsingElement* this, ParsingContext* context) {
 		}
 		// FIXME: Make sure it is the length and not the end offset
 		result           = Match_Success(vector[1], this, context);
-		DEBUG("[✓] %s:%s matched at %zd", this->name, config->expr, context->iterator->offset);
+		LOG_IF(context->grammar->isVerbose, "[✓] %s:%s matched at %zd", this->name, config->expr, context->iterator->offset);
 
 		// We create the token match
 		__ALLOC(TokenMatch, data);
@@ -714,7 +735,8 @@ Match* Token_recognize(ParsingElement* this, ParsingContext* context) {
 		context->iterator->move(context->iterator,result->length);
 		assert(Match_isSuccess(result));
 	}
-	return result;
+#endif
+	return MATCH_STATS(result);
 }
 
 const char* TokenMatch_group(Match* match, int index) {
@@ -733,11 +755,13 @@ void TokenMatch_free(Match* match) {
 	assert (match->context       != NULL);
 	assert (((ParsingElement*)(match->element))->type == TYPE_TOKEN);
 	TokenMatch* m = (TokenMatch*)match->data;
+#ifdef WITH_PCRE
 	if (m != NULL ) {
 		for (int j=0 ; j<m->count ; j++) {
 			pcre_free_substring(m->groups[j]);
 		}
 	}
+#endif
 }
 
 // ----------------------------------------------------------------------------
@@ -754,7 +778,7 @@ ParsingElement* Group_new(Reference* children[]) {
 }
 
 Match* Group_recognize(ParsingElement* this, ParsingContext* context){
-	DEBUGIF(strcmp(this->name, "_") != 0,"--- Group:%s at %zd", this->name, context->iterator->offset);
+	LOG_IF(context->grammar->isVerbose && strcmp(this->name, "_") != 0,"--- Group:%s at %zd", this->name, context->iterator->offset);
 	Reference* child = this->children;
 	Match*     match = NULL;
 	size_t     offset = context->iterator->offset;
@@ -764,7 +788,7 @@ Match* Group_recognize(ParsingElement* this, ParsingContext* context){
 			// The first succeding child wins
 			Match* result = Match_Success(match->length, this, context);
 			result->child = match;
-			return result;
+			return MATCH_STATS(result);
 		} else {
 			// Otherwise we skip to the next child
 			child = child->next;
@@ -772,11 +796,11 @@ Match* Group_recognize(ParsingElement* this, ParsingContext* context){
 	}
 	// If no child has succeeded, the whole group fails
 	if (context->iterator->offset != offset ) {
-		DEBUGIF(strcmp(this->name, "_") != 0,"[!] %s failed backtracking to %zd", this->name, offset);
+		LOG_IF(context->grammar->isVerbose && strcmp(this->name, "_") != 0,"[!] %s failed backtracking to %zd", this->name, offset);
 		Iterator_moveTo(context->iterator, offset);
 		assert( context->iterator->offset == offset );
 	}
-	return FAILURE;
+	return MATCH_STATS(FAILURE);
 }
 
 // ----------------------------------------------------------------------------
@@ -793,7 +817,7 @@ ParsingElement* Rule_new(Reference* children[]) {
 }
 
 Match* Rule_recognize (ParsingElement* this, ParsingContext* context){
-	DEBUGIF(strcmp(this->name, "_") != 0, "--- Rule:%s at %zd", this->name, context->iterator->offset);
+	LOG_IF(context->grammar->isVerbose && strcmp(this->name, "_") != 0, "--- Rule:%s at %zd", this->name, context->iterator->offset);
 	Reference* child  = this->children;
 	Match*     result = NULL;
 	Match*     last   = NULL;
@@ -827,8 +851,8 @@ Match* Rule_recognize (ParsingElement* this, ParsingContext* context){
 		child = child->next;
 		step++;
 	}
-	DEBUGIF( offset != context->iterator->offset && strcmp(this->name, "_") != 0 && !Match_isSuccess(result), "[!] %s[%d] failed at %zd", this->name, step, context->iterator->offset)
-	DEBUGIF( strcmp(this->name, "_") != 0 &&  Match_isSuccess(result), "[✓] %s[%d] matched at %zd", this->name, step, context->iterator->offset)
+	LOG_IF( context->grammar->isVerbose && offset != context->iterator->offset && strcmp(this->name, "_") != 0 && !Match_isSuccess(result), "[!] %s[%d] failed at %zd", this->name, step, context->iterator->offset)
+	LOG_IF( context->grammar->isVerbose && strcmp(this->name, "_") != 0 &&  Match_isSuccess(result), "[✓] %s[%d] matched at %zd", this->name, step, context->iterator->offset)
 	if (!Match_isSuccess(result)) {
 		// Match_free(result);
 		// If we had a failure, then we backtrack the iterator
@@ -842,7 +866,7 @@ Match* Rule_recognize (ParsingElement* this, ParsingContext* context){
 		// match.
 		result->length = last->offset - result->offset + last->length;
 	}
-	return result;
+	return MATCH_STATS(result);
 }
 
 // ----------------------------------------------------------------------------
@@ -863,7 +887,7 @@ Match*  Procedure_recognize(ParsingElement* this, ParsingContext* context) {
 	if (this->config != NULL) {
 		((ProcedureCallback)(this->config))(this, context);
 	}
-	return Match_Success(0, this, context);
+	return MATCH_STATS(Match_Success(0, this, context));
 }
 
 // ----------------------------------------------------------------------------
@@ -883,14 +907,14 @@ ParsingElement* Condition_new(ConditionCallback c) {
 Match*  Condition_recognize(ParsingElement* this, ParsingContext* context) {
 	if (this->config != NULL) {
 		Match* result = ((ConditionCallback)this->config)(this, context);
-		DEBUGIF(Match_isSuccess(result),  "[✓] Condition %s matched at %zd", this->name, context->iterator->offset)
-		DEBUGIF(!Match_isSuccess(result), "[1] Condition %s failed at %zd",  this->name, context->iterator->offset)
-		return  result;
+		LOG_IF(context->grammar->isVerbose &&  Match_isSuccess(result), "[✓] Condition %s matched at %zd", this->name, context->iterator->offset)
+		LOG_IF(context->grammar->isVerbose && !Match_isSuccess(result), "[1] Condition %s failed at %zd",  this->name, context->iterator->offset)
+		return  MATCH_STATS(result);
 	} else {
-		DEBUGIF("[✓] Condition %s matched by default at %zd", this->name, context->iterator->offset)
+		LOG_IF(context->grammar->isVerbose, "[✓] Condition %s matched by default at %zd", this->name, context->iterator->offset)
 		Match* result = Match_Success(0, this, context);
 		assert(Match_isSuccess(result));
-		return  result;
+		return  MATCH_STATS(result);
 	}
 }
 
@@ -922,7 +946,6 @@ void ParsingStep_free( ParsingStep* this ) {
 //
 // ----------------------------------------------------------------------------
 
-
 ParsingOffset* ParsingOffset_new( size_t offset ) {
 	__ALLOC(ParsingOffset, this);
 	this->offset = offset;
@@ -942,13 +965,119 @@ void ParsingOffset_free( ParsingOffset* this ) {
 	__DEALLOC(this);
 }
 
+// ----------------------------------------------------------------------------
+//
+// PARSING CONTEXT
+//
+// ----------------------------------------------------------------------------
+
+ParsingContext* ParsingContext_new( Grammar* g, Iterator* iterator ) {
+	__ALLOC(ParsingContext, this);
+	assert(g);
+	assert(iterator);
+	this->grammar  = g;
+	this->iterator = iterator;
+	this->stats    = ParsingStats_new();
+	ParsingStats_setSymbolsCount(this->stats, g->axiomCount + g->skipCount);
+	this->offsets  = NULL;
+	this->current  = NULL;
+	return this;
+}
+
+void ParsingContext_free( ParsingContext* this ) {
+	if (this!=NULL) {
+		Iterator_free(this->iterator);
+		ParsingStats_free(this->stats);
+		__DEALLOC(this);
+	}
+}
+
+// ----------------------------------------------------------------------------
+//
+// PARSING STATS
+//
+// ----------------------------------------------------------------------------
+
+ParsingStats* ParsingStats_new(void) {
+	__ALLOC(ParsingStats,this);
+	this->bytesRead = 0;
+	this->parseTime = 0;
+	this->successBySymbol = NULL;
+	this->failureBySymbol = NULL;
+	return this;
+}
+
+void ParsingStats_free(ParsingStats* this) {
+	if (this != NULL) {
+		__DEALLOC(this->successBySymbol);
+		__DEALLOC(this->failureBySymbol);
+	}
+	__DEALLOC(this);
+}
+
+void ParsingStats_setSymbolsCount(ParsingStats* this, size_t t) {
+	__DEALLOC(this->successBySymbol);
+	__DEALLOC(this->failureBySymbol);
+	this->successBySymbol = calloc(sizeof(size_t), t);
+	this->failureBySymbol = calloc(sizeof(size_t), t);
+	this->symbolsCount    = t;
+}
+
+Match* ParsingStats_registerMatch(ParsingStats* this, Element* e, Match* m) {
+	// We can convert ParsingElements to Reference and vice-versa as they
+	// have the same start sequence (cart type, int id).
+	Reference* r = (Reference*)e;
+	if (Match_isSuccess(m)) {
+		this->successBySymbol[r->id] += 1;
+	} else {
+		this->failureBySymbol[r->id] += 1;
+	}
+	return m;
+}
+
+// ----------------------------------------------------------------------------
+//
+// PARSING RESULT
+//
+// ----------------------------------------------------------------------------
+
+ParsingResult* ParsingResult_new(Match* match, ParsingContext* context) {
+	__ALLOC(ParsingResult,this);
+	assert(match != NULL);
+	assert(context != NULL);
+	assert(context->iterator != NULL);
+	this->match   = match;
+	this->context = context;
+	if (match != FAILURE) {
+		if (Iterator_hasMore(context->iterator) && Iterator_remaining(context->iterator) > 0) {
+			LOG_IF(context->grammar->isVerbose, "Partial success, parsed %zd bytes, %zd remaining", context->iterator->offset, Iterator_remaining(context->iterator));
+			this->status = STATUS_PARTIAL;
+		} else {
+			LOG_IF(context->grammar->isVerbose, "Succeeded, parsed %zd bytes", context->iterator->offset);
+			this->status = STATUS_SUCCESS;
+		}
+	} else {
+		LOG_IF(context->grammar->isVerbose, "Failed, parsed %zd bytes, %zd remaining", context->iterator->offset, Iterator_remaining(context->iterator))
+		this->status = STATUS_FAILED;
+	}
+	return this;
+}
+
+void ParsingResult_free(ParsingResult* this) {
+	if (this != NULL) {
+		Match_free(this->match);
+		ParsingContext_free(this->context);
+	}
+	__DEALLOC(this);
+}
+
 // ----------------------------------------------------------------------------
 //
 // GRAMMAR
 //
 // ----------------------------------------------------------------------------
 
-int Grammar__resetElementIDs(Element* e, int step) {
+int Grammar__resetElementIDs(Element* e, int step, void* nothing) {
 	if (Reference_Is(e)) {
 		Reference* r = (Reference*)e;
 		if (r->id != ID_BINDING) {
@@ -968,12 +1097,12 @@ int Grammar__resetElementIDs(Element* e, int step) {
 	}
 }
 
-int Grammar__assignElementIDs(Element* e, int step) {
+int Grammar__assignElementIDs(Element* e, int step, void* nothing) {
 	if (Reference_Is(e)) {
 		Reference* r = (Reference*)e;
 		if (r->id == -1) {
 			r->id = step;
-			DEBUGIF(r->name != NULL, "[%03d] [%c] %s", r->id, r->type, r->name);
+			DEBUG_IF(r->name != NULL, "[%03d] [%c] %s", r->id, r->type, r->name);
 			return step;
 		} else {
 			return -1;
@@ -982,7 +1111,7 @@ int Grammar__assignElementIDs(Element* e, int step) {
 		ParsingElement * r = (ParsingElement*)e;
 		if (r->id == -1) {
 			r->id = step;
-			DEBUGIF(r->name != NULL, "[%03d] [%c] %s", r->id, r->type, r->name);
+			DEBUG_IF(r->name != NULL, "[%03d] [%c] %s", r->id, r->type, r->name);
 			return step;
 		} else {
 			return -1;
@@ -990,63 +1119,67 @@ int Grammar__assignElementIDs(Element* e, int step) {
 	}
 }
 
+int Grammar__registerElement(Element* e, int step, void* grammar) {
+	Reference* r  = (Reference*)e;
+	Grammar*   g  = (Grammar*)grammar;
+	Element*   ge = g->elements[r->id];
+	if (ge == NULL) {
+		g->elements[r->id] = e;
+		return step;
+	} else {
+		return -1;
+	}
+}
+
 void Grammar_prepare ( Grammar* this ) {
 	if (this->skip!=NULL)  {
 		this->skip->id = 0;
 	}
 	if (this->axiom!=NULL) {
-		Element_walk(this->axiom, Grammar__resetElementIDs);
+		Element_walk(this->axiom, Grammar__resetElementIDs, NULL);
 		if (this->skip != NULL) {
-			Element_walk(this->skip, Grammar__resetElementIDs);
+			Element_walk(this->skip, Grammar__resetElementIDs, NULL);
 		}
-		int count = Element_walk(this->axiom, Grammar__assignElementIDs);
+		int count = Element_walk(this->axiom, Grammar__assignElementIDs, NULL);
 		this->axiomCount = count;
 		if (this->skip != NULL) {
-			this->skipCount = Element__walk(this->skip, Grammar__assignElementIDs, count) - count;
+			this->skipCount = Element__walk(this->skip, Grammar__assignElementIDs, count, NULL) - count;
+		}
+		// Now we register the elements
+		__DEALLOC(this->elements);
+		this->elements = calloc( sizeof(Element*), this->skipCount + this->axiomCount);
+		count = Element_walk(this->axiom, Grammar__registerElement, this);
+		if (this->skip != NULL) {
+			Element__walk(this->skip, Grammar__registerElement, count, this);
 		}
 	}
 }
 
-Match* Grammar_parseFromIterator( Grammar* this, Iterator* iterator ) {
+ParsingResult* Grammar_parseFromIterator( Grammar* this, Iterator* iterator ) {
 	assert(this->axiom != NULL);
-	ParsingOffset* offset  = ParsingOffset_new(iterator->offset);
-	ParsingContext context = (ParsingContext){
-		.grammar  = this,
-		.iterator = iterator,
-		.offsets  = offset,
-		.offsets  = offset,
-	};
-	Match* match = this->axiom->recognize(this->axiom, &context);
-	if (match != FAILURE) {
-		if (Iterator_hasMore(context.iterator) && Iterator_remaining(context.iterator) > 0) {
-			DEBUG("Partial success, parsed %zd bytes, %zd remaining", context.iterator->offset, Iterator_remaining(context.iterator));
-		} else {
-			DEBUG("Succeeded, parsed %zd bytes", context.iterator->offset);
-		}
-	} else {
-		DEBUG("Failed, parsed %zd bytes, %zd remaining", context.iterator->offset, Iterator_remaining(context.iterator))
-	}
-	return match;
+	// ParsingOffset*  offset  = ParsingOffset_new(iterator->offset);
+	ParsingContext* context = ParsingContext_new(this, iterator);
+	clock_t t1  = clock();
+	Match* match = this->axiom->recognize(this->axiom, context);
+	context->stats->parseTime = ((double)clock() - (double)t1) / CLOCKS_PER_SEC;
+	context->stats->bytesRead = iterator->offset;
+	return ParsingResult_new(match, context);
 }
 
-Match* Grammar_parseFromPath( Grammar* this, const char* path ) {
+ParsingResult* Grammar_parseFromPath( Grammar* this, const char* path ) {
 	Iterator* iterator = Iterator_Open(path);
 	if (iterator != NULL) {
-		Match*    result   = Grammar_parseFromIterator(this, iterator);
-		Iterator_free(iterator);
-		return result;
+		return Grammar_parseFromIterator(this, iterator);
 	} else {
 		errno = ENOENT;
 		return NULL;
 	}
 }
 
-Match* Grammar_parseFromString( Grammar* this, const char* text ) {
+ParsingResult* Grammar_parseFromString( Grammar* this, const char* text ) {
 	Iterator* iterator = Iterator_FromString(text);
 	if (iterator != NULL) {
-		Match*    result   = Grammar_parseFromIterator(this, iterator);
-		Iterator_free(iterator);
-		return result;
+		return Grammar_parseFromIterator(this, iterator);
 	} else {
 		errno = ENOENT;
 		return NULL;
